#ifndef POOLCORE_CONTEXT_H
#define POOLCORE_CONTEXT_H

#include "POOLCore/Properties.h"
#include "RelationalAccess/AccessMode.h"
#include <string>
#include <map>

namespace coral {
  class IConnectionService;
  class ISessionProxy;
}

namespace pool {

  class IBlobStreamingService;

  /// Class defining the entry point to the framework component model for POOL.
  class Context {

    public:

    static std::string automaticSchemaEvolutionFlagName();
    
    public:

    Context();

    virtual ~Context();

    void setBlobStreamer(pool::IBlobStreamingService* service, bool giveUpOwnership=true);
    
    pool::IBlobStreamingService* blobStreamer();

    void setConnectionService(coral::IConnectionService* service, bool giveUpOwnership=true);

    coral::IConnectionService* connectionService();

    void enableSessionSharing();

    void disableSessionSharing();

    bool isSessionSharingEnabled();

    Properties& properties();
    
    coral::ISessionProxy& sharedSession( const std::string& connectionString, coral::AccessMode accessMode );

    private:

    coral::ISessionProxy* getSession( const std::string& connectionString,
                                      coral::AccessMode accessMode,
                                      std::map<std::string,coral::ISessionProxy*>& sessionPool);

    private:

    pool::IBlobStreamingService* m_blobStreamingService;

    bool m_blobStreamingServiceOwnership;

    coral::IConnectionService* m_connectionService;

    bool m_connectionServiceOwnership;

    std::map<std::string,coral::ISessionProxy*> m_readOnlySessions;
    std::map<std::string,coral::ISessionProxy*> m_updateSessions;

    bool m_sessionSharing;

    Properties m_properties;
    
  };

}

inline
coral::IConnectionService*
pool::Context::connectionService(){
  return m_connectionService;
}

inline
pool::IBlobStreamingService*
pool::Context::blobStreamer(){
  return m_blobStreamingService;
}

inline
void 
pool::Context::enableSessionSharing(){
  m_sessionSharing = true;
}

inline
void 
pool::Context::disableSessionSharing(){
  m_sessionSharing = false;
}

inline
bool 
pool::Context::isSessionSharingEnabled(){
  return m_sessionSharing;
}

inline
pool::Properties&
pool::Context::properties(){
  return m_properties;
}



#endif
