#ifndef POOL_PCONTAINERIMPL_H
#define POOL_PCONTAINERIMPL_H

template <class Tp> pool::Range<Tp>::Range():PContainerBase(),m_data(new PContainerData<Tp>){
  m_data->addRef();
}

template <class Tp> pool::Range<Tp>::Range(pool::PContainerData<Tp>& data):PContainerBase(),m_data(&data){
  m_data->addRef();
}

template <class Tp> pool::Range<Tp>::Range(const pool::Range<Tp>& rhs):PContainerBase(),m_data(rhs.m_data){
  m_data->addRef();
}

template <class Tp> pool::Range<Tp>::~Range(){
  releaseData(m_data);
}
      
template <class Tp> pool::Range<Tp>& pool::Range<Tp>::operator=(const pool::Range<Tp>& rhs){
  if(&rhs != this){
    PContainerData<Tp>* prevData = m_data;
    m_data = rhs.m_data;
    if(m_data) m_data->addRef();
    releaseData( prevData );
  }
  return *this;
}

template <class Tp> typename pool::Range<Tp>::const_iterator pool::Range<Tp>::begin() const {
  return m_data->cbegin();
}

template <class Tp> typename pool::Range<Tp>::const_iterator pool::Range<Tp>::end() const {
  return m_data->cend();
}

template <class Tp> typename pool::Range<Tp>::const_reverse_iterator pool::Range<Tp>::rbegin() const {
  return m_data->crbegin();
}

template <class Tp> typename pool::Range<Tp>::const_reverse_iterator pool::Range<Tp>::rend() const {
  return m_data->crend();
}

template <class Tp> size_t pool::Range<Tp>::size() const {
  return m_data->size();
}

template <class Tp> size_t pool::Range<Tp>::frontIndex() const {
  return m_data->front().first;
}

template <class Tp> size_t pool::Range<Tp>::backIndex() const {
  return m_data->back().first;
}



template <class Tp> pool::Query<Tp>::Query(pool::IContainerLoader& loader):LoaderClient(loader){
}

template <class Tp>
template <typename Prim> void pool::Query<Tp>::addSelection(const std::string& dataMemberName, pool::SelectionItemType stype, Prim selectionData){
  m_selection.addDataItem(dataMemberName, stype, selectionData);
}

template <class Tp> size_t pool::Query<Tp>::count(){
  return loader().getSelectionCount( m_selection );
}

template <class Tp> pool::Range<Tp> pool::Query<Tp>::execute(){
  PContainerData<Tp>* newData = new PContainerData<Tp>;
  loader().loadSelection( m_selection, const_cast<void*>(newData->storageAddress()) );
  return Range<Tp>(*newData);
}

template <class Tp> pool::PContainer<Tp>::PContainer():PContainerBase(),LoaderClient(),m_data(new PContainerData<Tp>),m_isLocked(false),m_isLoaded(false){
  m_data->addRef();
}
    
template <class Tp> pool::PContainer<Tp>::PContainer(size_t n, const Tp& value):PContainerBase(),LoaderClient(),m_data(new PContainerData<Tp>(n,value)),m_isLocked(false),m_isLoaded(false){
  m_data->addRef();
}
    
template <class Tp> pool::PContainer<Tp>::PContainer(const PContainer<Tp>& rhs):PContainerBase(),LoaderClient(rhs),m_data(rhs.m_data),m_isLocked(rhs.m_isLocked),m_isLoaded(rhs.m_isLoaded){
  m_data->addRef();
}

template <class Tp> pool::PContainer<Tp>::~PContainer(){
  releaseData(m_data);
}

template <class Tp> pool::PContainer<Tp>& pool::PContainer<Tp>::operator=(const pool::PContainer<Tp>& rhs){
  if(&rhs != this){
    PContainerData<Tp>* prevData = m_data;
    m_data = rhs.m_data;
    if(m_data) m_data->addRef();
    releaseData( prevData );
    m_isLocked = rhs.m_isLocked;
    m_isLoaded = rhs.m_isLoaded;
  }
  return *this;
}

template <class Tp> pool::Range<Tp> pool::PContainer<Tp>::select(int startIndex, int endIndex) const {
  if(m_isLocked ){
    except("The Container is locked in writing mode, cannot make queries.");
  }
  Selection sel;
  sel.addIndexItem( startIndex, endIndex );
  if(!hasLoader()){
    except("The Loader is not installed.");
  }
  PContainerData<Tp>* newData = new PContainerData<Tp>;
  loader().loadSelection( sel, const_cast<void*>(newData->storageAddress()) );
  return Range<Tp>(*newData);
}
    
template <class Tp> pool::Range<Tp> pool::PContainer<Tp>::select(const pool::Selection& sel) const {
  if(m_isLocked ){
    except("The Container is locked in writing mode, cannot make queries.");
  }
  if(!hasLoader()){
    except("The Loader is not installed.");
  }
  PContainerData<Tp>* newData = new PContainerData<Tp>;
  loader().loadSelection( sel, const_cast<void*>(newData->storageAddress()) );
  return Range<Tp>(*newData);
}

template <class Tp> pool::Query<Tp> pool::PContainer<Tp>::query() const{
  if(m_isLocked ){
    except("The Container is locked in writing mode, cannot make queries.");
  }
  if(!hasLoader()){
    except("The Loader is not installed.");
  }
  return Query<Tp>(loader());
}
    
template <class Tp> bool pool::PContainer<Tp>::lock() {
  bool wasLocked = m_isLocked;
  m_isLocked = true;
  return wasLocked;
}
        
template <class Tp> bool pool::PContainer<Tp>::isLocked() const {
  return m_isLocked;
}

template <class Tp> typename pool::PContainer<Tp>::iterator pool::PContainer<Tp>::begin(){
  initialize();
  return m_data->begin();
}

template <class Tp> typename pool::PContainer<Tp>::iterator pool::PContainer<Tp>::end(){
  initialize();
  return m_data->end();
}

template <class Tp> typename pool::PContainer<Tp>::const_iterator pool::PContainer<Tp>::begin() const {
  initialize();
  return m_data->cbegin();
}

template <class Tp> typename pool::PContainer<Tp>::const_iterator pool::PContainer<Tp>::end() const {
  initialize();
  return m_data->cend();
}

template <class Tp> typename pool::PContainer<Tp>::reverse_iterator pool::PContainer<Tp>::rbegin(){
  initialize();
  return m_data->rbegin();
}
    
template <class Tp> typename pool::PContainer<Tp>::reverse_iterator pool::PContainer<Tp>::rend(){
  initialize();
  return m_data->rend();
}
    
template <class Tp> typename pool::PContainer<Tp>::const_reverse_iterator pool::PContainer<Tp>::rbegin() const {
  initialize();
  return m_data->crbegin();
}
    
template <class Tp> typename pool::PContainer<Tp>::const_reverse_iterator pool::PContainer<Tp>::rend() const {
  return m_data->crend();
}

template <class Tp> size_t pool::PContainer<Tp>::size() const {
  initialize();
  return m_data->size();
}

template <class Tp> size_t pool::PContainer<Tp>::max_size() const {
  return m_data->max_size();
}
    
template <class Tp> void pool::PContainer<Tp>::resize(size_t n, const Tp& value){
  initialize();
  m_data->resize(n,value);
}
    
template <class Tp> size_t pool::PContainer<Tp>::capacity() const {
  return m_data->capacity();
}
    
template <class Tp> bool pool::PContainer<Tp>::empty() const {
  initialize();
  return m_data->empty();
}
    
template <class Tp> void pool::PContainer<Tp>::reserve(size_t n) {
  m_data->reserve(n);
}

template <class Tp> typename pool::PContainer<Tp>::reference pool::PContainer<Tp>::operator[] ( size_t n ){
  initialize();
  return m_data->operator[](n);
}

template <class Tp> typename pool::PContainer<Tp>::const_reference pool::PContainer<Tp>::operator[] ( size_t n ) const {
  initialize();
  return m_data->operator[](n);
}

template <class Tp> typename pool::PContainer<Tp>::const_reference pool::PContainer<Tp>::at( size_t n ) const {
  initialize();
  return m_data->operator[](n);
}
    
template <class Tp> typename pool::PContainer<Tp>::reference pool::PContainer<Tp>::at( size_t n ) {
  initialize();
  return m_data->operator[](n);
}

template <class Tp> typename pool::PContainer<Tp>::reference pool::PContainer<Tp>::front ( ) {
  return m_data->front();
}
    
template <class Tp> typename pool::PContainer<Tp>::const_reference pool::PContainer<Tp>::front ( ) const {
  return m_data->front();
}
    
template <class Tp> typename pool::PContainer<Tp>::reference pool::PContainer<Tp>::back ( ) {
  return m_data->back();
}
    
template <class Tp> typename pool::PContainer<Tp>::const_reference pool::PContainer<Tp>::back ( ) const {
  return m_data->back();
}

template <class Tp> void pool::PContainer<Tp>::assign ( size_t n, const Tp& u ) {
  m_data->assign(n,u);
}

template <class Tp> void pool::PContainer<Tp>::push_back ( const Tp& x ){
  initialize();
  m_isLocked = true;
  m_data->push_back(x);
}

template <class Tp> void pool::PContainer<Tp>::pop_back (){
  initialize();
  m_isLocked = true;
  m_data->pop_back();
}

template <class Tp> void pool::PContainer<Tp>::clear ( ){
  m_data->clear();
  m_isLoaded = false;
}

template <class Tp> void pool::PContainer<Tp>::reset ( ){
  m_data->clear();
  m_isLoaded = false;
  m_isLocked = false;
}

template <class Tp> bool pool::PContainer<Tp>::operator==(const pool::PContainer<Tp>& vec) const {
  initialize();
  vec.initialize();
  return m_data->operator==(*vec.m_data);
}

template <class Tp> bool pool::PContainer<Tp>::operator!=(const pool::PContainer<Tp>& vec) const {
  initialize();
  vec.initialize();
  return m_data->operator!=(*vec.m_data);
}

template <class Tp> void pool::PContainer<Tp>::installLoader(pool::IContainerLoader& loader){
  LoaderClient::install(loader);
}

template <class Tp> size_t pool::PContainer<Tp>::persistentSize() const {
  return m_data->persistentSize();
}

template <class Tp> const void* pool::PContainer<Tp>::storageAddress() const {
  return m_data->storageAddress();
}

template <class Tp> void pool::PContainer<Tp>::initialize() const {
  if(hasLoader() && !m_isLocked && !m_isLoaded){
    loader().load(const_cast<void*>(m_data->storageAddress()));
    m_isLoaded = true;
  }
}

#endif  // POOL_PCONTAINERIMPL_H
