#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMWriter.hpp>
#include <xercesc/dom/DOMNode.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>

#include <xercesc/sax/SAXException.hpp>

#include <xercesc/framework/StdOutFormatTarget.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>
#include <xercesc/framework/StdOutFormatTarget.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>

#include <xercesc/util/XMLUni.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLUni.hpp>
#include <xercesc/util/OutOfMemoryException.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/TransService.hpp>

//#include <xercesc/parsers/XercesDOMParser.hpp>

#include <ctime>
#include <stdexcept>

#include "ArgQual.h"
#include "CmdLineArgs2.h"
#include "Args2Container.h"

XERCES_CPP_NAMESPACE_USE

// Stole this little snippet from OOPWeb.com 
// http://oopweb.com/CPP/Documents/CPPHOWTO/Volume/C++Programming-HOWTO-7.html
// to fill in a basic failure in std::string 
// Essentially it does for C++ what strtok does for C

void Tokenize(const std::string& str,
              std::vector<std::string>& tokens,
              const std::string& delimiters = " ")
{
    // Skip delimiters at beginning.
    std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    std::string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (std::string::npos != pos || std::string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

/***************************************************************

Args2Container is an extension of a vector of CmdLineArgs2

It has a name and a desc. Name is normally the program name.
It also adds several utility methods such as

printHelp(verbose) :
   - (verbose = false) print out names of ArgQuals 
     for all contained pointers
   - (verbose = true) print out names _and_ desc of ArgQuals 
     for all contained pointers

checkValid() : Loop over pointers and call valid() method

evalArgs(argv) : save time/errors by calling evalArgs 
   on the container rather than the individual CmdLineArg2's

***************************************************************/
Args2Container::Args2Container( std::string name, bool doxml, coral::MessageStream *l )
      : m_name(name),
        m_argc(0),
        m_xml(doxml),
        m_xmlIn(DISCARD),
	m_ver( V2 ),
        m_a2c_cla2("local"),
        log( l? *l : *new coral::MessageStream("CmdLine") )
{
   QualList markers;
   markers.insert(std::make_pair("-xmlInput",ArgQual()) );
   markers.insert(std::make_pair("-compatibility",ArgQual()) );
   markers["-xmlInput"].desc << "This takes a file name as input. " << std::endl
                             << "NOTE: The CLI args will be discarded! ";
   markers["-compatibility"].desc << "If this is set, then it uses the old xml format" << std::endl 
                                  << "  Options are : "; 
   for (unsigned int i=0; i<NUM_VERSIONS; ++i) { 
      markers["-compatibility"].desc << " V" << i+1; 
   }
   m_a2c_cla2.setArgQuals(markers);
   this->push_back(&m_a2c_cla2);
}

void Args2Container::init()
{
   this->clear();
   m_argc=0;
   m_xml=false;
   m_xmlIn=DISCARD;
}

//-----------------------------------------------------------------------
// printHelp : loop over contents and print name (and desc) of qualifiers
//-----------------------------------------------------------------------
void Args2Container::printHelp(bool verbose) const
{
   std::cerr << "Usage: " << m_name << std::endl;
   if (desc.str().size()>0) std::cerr << desc.str() << std::endl << std::endl; 
   for (Args2Container::const_iterator cmdit = this->begin(); cmdit != this->end(); ++cmdit)
   {
      for (QualList::const_iterator it = (*cmdit)->getArgQuals().begin();
                                    it != (*cmdit)->getArgQuals().end();
                                  ++it)
      {
         if (it->first != "-help") {
            std::cerr << "Option = " << it->first << " ";
            if ( verbose ) it->second.print(std::cout);
            std::cerr << std::endl;
         }
      }
   }
}

bool Args2Container::checkValid() const
{
   for (Args2Container::const_iterator ait = this->begin(); ait != this->end(); ++ait) {
      if ( (*ait)->valid() == false ) {
         std::cerr << "Invalid Qualifier" << std::endl;
         return false;
      }
   }
   return true;
}

//
// Evaluate the argv vector for a named argument
//
bool Args2Container::evalArgs(const std::string cliarg, std::vector<std::string>& argv)
{
   std::string file;
   // Evaluate the args for the Args2Container local cla2
   bool test = m_a2c_cla2.evalArgs(argv);
   // If this is OK then check if xmlInput is one of the options
   if (test) {
      CmdLineArgs2::iterator clit = m_a2c_cla2.begin();
      while (clit!=m_a2c_cla2.end()) {
         if (clit->first == "-xmlInput") {
            file = argv[clit->second.first];
            std::cout << "Taking input from xml file" << std::endl;
            argv = this->fillCmdLineArgsFromXML(file);
         }
	 if (clit->first == "-compatibility") { 
	    if (argv[clit->second.first]=="V1") m_ver=V1; 
	    else if (argv[clit->second.first]=="V2") m_ver=V2; 
	    else {std::cout << "compatibility version " << argv[clit->second.first]  
			    << " not supported" << std::endl;} 
	 }
	 ++clit;
      }
   }
   else { 
      std::cerr << "Args2Container has invalid internal argument" << std::endl;
   } 

   // If Args2Container is valid, then eval args for all content cla2's
   for (Args2Container::const_iterator ait = this->begin(); ait != this->end(); ++ait) {
      if ( cliarg=="ALL" || (*ait)->name()==cliarg) {
         if ( (*ait)->evalArgs(argv) == false ) {
            std::cerr << "Unable to eval args for " << (*ait)->name() << std::endl;
            return false;
         }
      }
   }
   m_argc=argv.size();

   // If output xml is requested ...
   // Note that it always contains what was finally evaluated, NOT just the command line
   if (doXML()) {
      try {
         XMLPlatformUtils::Initialize();
      }
      catch( const XMLException& ) {
         std::cerr << "Error during initialization! :\n" << std::endl;
      }
      DOMNode* doc = 0;
      doc = defineXMLContent(argv);
      if (doc) writeXML(doc);
      else std::cerr << "Unable to write XML" << std::endl;
   }
   return true;
}

//
// Evaluate the argv vector for ALL contents
//
bool Args2Container::evalArgs(std::vector<std::string>& argv)
{
   return this->evalArgs("ALL",argv);
}




#include <boost/scoped_array.hpp>

typedef std::basic_string<XMLCh> XercesString;

// Converts from a narrow-character string to a wide-character string.
inline XercesString fromNative(const char* str)
{
    boost::scoped_array<XMLCh> ptr( XMLString::transcode(str) );
    return XercesString(ptr.get( ));
}

// Converts from a narrow-character string to a wide-charactr string.
inline XercesString fromNative(const std::string& str)
{
    return fromNative(str.c_str( ));
}

// Converts from a wide-character string to a narrow-character string.
inline std::string toNative(const XMLCh* str)
{
   if( !str ) return "";
   boost::scoped_array<char> ptr( XMLString::transcode(str) );
   return std::string(ptr.get( ));
}

// Converts from a wide-character string to a narrow-character string.
inline std::string toNative(const XercesString& str)
{
    return toNative(str.c_str( ));
}





DOMNode* Args2Container::defineXMLContent(std::vector<std::string>& argv)
{
   // Initialaize to zero
   DOMDocument* newDocument = 0;
   try {
      // Create empty document based on Args2Container.dtd
      const XMLCh ls_id [] = {chLatin_L, chLatin_S, chNull};
      DOMImplementation* impl(DOMImplementationRegistry::getDOMImplementation (ls_id));
      DOMDocumentType* Arg2Type = impl->createDocumentType(XMLString::transcode("ArgList"),
                                                           XMLString::transcode(""),
                                                           XMLString::transcode("") );
      newDocument = impl->createDocument(0, XMLString::transcode("ArgList"), Arg2Type);

      // Create the base element, assign the id, and add to tree
      DOMElement* toolBase = newDocument->createElement(XMLString::transcode("ToolInfo"));
      toolBase->setAttribute(XMLString::transcode("toolID"),
                             XMLString::transcode(m_name.c_str()));
      time_t tm = time(NULL);
      char * date = asctime(localtime(&tm));
      toolBase->setAttribute(XMLString::transcode("date"),
                             XMLString::transcode(date));
      newDocument->getDocumentElement()->appendChild(toolBase);

      // Loop over container contents
      for (Args2Container::const_iterator cmdit = this->begin(); cmdit != this->end(); ++cmdit)
      {
         // Loop over ArgQual's in each CLA2
         for (QualList::const_iterator it = (*cmdit)->getArgQuals().begin();
                                       it != (*cmdit)->getArgQuals().end();
                                     ++it)
         {
            // Get the pointer to the CLA2
            CmdLineArgs2* ptr = (*cmdit);
	    if (ptr->name() != "local") { 
	       // Get pointer to the data for the qualifier 
	       CmdLineArgs2::const_iterator ptr_it = ptr->find(it->first); 
	       int first=0,last=0; 
	       // Second part of iterator is pair of endpoints for the data in the argv 
	       if (ptr_it != ptr->end()) {first=ptr_it->second.first;last=ptr_it->second.second;} 
	       else continue; 
	       // Fill a stringstream with everything in the argv for this qualifier 
	       std::stringstream argString; 
	       for (int j=first; j<last; ++j) {  
		  argString << argv[j]; 
		  if (j < (last - 1) ) argString << " "; 
	       } 
	       // Create an element for the CLA2 and append it to the toolBase 
	       DOMElement* cliArg = newDocument->createElement(XMLString::transcode("CliArg")); 
	       toolBase->appendChild(cliArg); 
	       // Then set the attribute data name,option 
	       //   and create a child text node with the data for that option 
	       cliArg->setAttribute(XMLString::transcode("name"), 
				    XMLString::transcode((*cmdit)->name().c_str())); 
	       cliArg->setAttribute(XMLString::transcode("option"), 
				    XMLString::transcode(it->first.c_str())); 
	       // Check whether argument takes data 
	       if (it->second.argsize) {  
		  if (m_ver==V1) { 
		     std::cout << "using compatibility for xml" << std::endl; 
		     cliArg->appendChild( 
			newDocument->createTextNode(XMLString::transcode(argString.str().c_str())) 
			); 
		  } 
		  else { 
		     for (int j=first; j<last; ++j) { 
			//std::cout << it->first << " : " << j << " : " << argv[j] << std::endl; 
			DOMElement* datum = newDocument->createElement(XMLString::transcode("datum")); 
			cliArg->appendChild(datum); 
			datum->appendChild( 
			   newDocument->createTextNode(XMLString::transcode(argv[j].c_str())) 
			   ); 
		     } 
		  } // version check 
	       }
	    }
	 } // QualList iterator 
      } // Args2Container iterator 	    
   }
   catch (...) {
      std::cerr << "Problem in defineXMLContent" << std::endl;
   }
   return newDocument;
}


void Args2Container::writeXML(DOMNode* newDocument)
{
   try {
      // get a writer instance
      const XMLCh ls_id [] = {chLatin_L, chLatin_S, chNull};
      DOMImplementation* impl(DOMImplementationRegistry::getDOMImplementation (ls_id));

      DOMWriter         *theSerializer = ((DOMImplementationLS*)impl)->createDOMWriter();

      if (theSerializer->canSetFeature(XMLUni::fgDOMWRTDiscardDefaultContent, true))
          theSerializer->setFeature(XMLUni::fgDOMWRTDiscardDefaultContent, true);

      if (theSerializer->canSetFeature(XMLUni::fgDOMWRTFormatPrettyPrint, true))
          theSerializer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint, true);

      if (theSerializer->canSetFeature(XMLUni::fgDOMWRTBOM, true))
          theSerializer->setFeature(XMLUni::fgDOMWRTBOM, true);

      // set a target as the file argument
      std::string file(m_name+".xml");
      log << coral::Debug << "About to write file " << file <<  coral::MessageStream::endmsg;
      XMLFormatTarget* myFormTarget = new LocalFileFormatTarget(file.c_str());

      // write document to target
      theSerializer->writeNode(myFormTarget, *newDocument);

      // clean up the mess
      if (theSerializer!=NULL && theSerializer!=0) delete theSerializer;
      if (myFormTarget!=NULL && myFormTarget!=0)   delete myFormTarget;
      if (newDocument!=NULL && newDocument!=0)     delete newDocument;
   }
   catch (const SAXException& e) {
        std::cout << "xml error: " << e.getMessage( ) << "\n";
   } 
   catch (const DOMException& e) {
        std::cout << "xml error: " << e.getMessage( ) << "\n";
   }
   catch( const XMLException& e ) {
      log << coral::Warning << "Failed writing XML options file: " << toNative(e.getMessage()) << coral::MessageStream::endmsg;
   }
}

//------------------------------------------------------------------------------
//
// fillCmdLineArgsFromXML
//
//   Given a file name it will try to read it as xml
//   and return a vector<string> which can be used to replace the argv_v.
//   DTD is implicit in code, but not checked explicitly.
//
//------------------------------------------------------------------------------
std::vector<std::string> Args2Container::fillCmdLineArgsFromXML(std::string file)
{
   std::vector<std::string> args;
   args.push_back(m_name);

   try {
      XMLPlatformUtils::Initialize();
   }
   catch( const XMLException& ) {
      std::cerr << "fillCmdLineArgsFromXML: Error during initialization!" << std::endl;
      return args;
   }

   XercesDOMParser *parser = new XercesDOMParser;
   parser->setValidationScheme(XercesDOMParser::Val_Auto);
   parser->setDoNamespaces(false);
   //DOMTreeErrorReporter *errReporter = new DOMTreeErrorReporter();
   //parser->setErrorHandler(errReporter);

   bool errorsOccured = false;
   try
   {
       parser->parse(file.c_str());
   }
   catch (const OutOfMemoryException&)
   {
      std::cerr << "OutOfMemoryException" << std::endl;
      errorsOccured = true;
   }
   catch (const XMLException& e)
   {
      std::cerr << "An error occurred during parsing\n   Message: "
                << e.getMessage() << std::endl;
      errorsOccured = true;
   }

   catch (const DOMException& e)
   {
      const unsigned int maxChars = 2047;
      XMLCh errText[maxChars + 1];

      std::cerr << "\nDOM Error during parsing: '" << file << "'\n"
                << "DOMException code is:  " << e.code << std::endl;

      if (DOMImplementation::loadDOMExceptionMsg(e.code, errText, maxChars))
          std::cerr << "Message is: " << errText << std::endl;

      errorsOccured = true;
   }

   catch (...)
   {
      std::cerr << "An error occurred during parsing\n " << std::endl;
      errorsOccured = true;
   }

   // if any errors just exit now 
   if (errorsOccured) return args;

   DOMDocument* doc = parser->getDocument();

   // Look for ToolInfo elements
   DOMNodeList* toolList = doc->getElementsByTagName(XMLString::transcode("ToolInfo"));
   // If there is not exactly one exit now
   if (toolList->getLength()!=1) {
      std::cout << "Wrong number of  ToolInfo records" << std::endl;
      return args;
   }
   // since there is only one, take the first item 
   DOMNode* toolInfo = toolList->item(0);
   if (toolInfo!=NULL) {
      DOMNamedNodeMap* toolAttrs=toolInfo->getAttributes();
      std::string toolID = std::string(XMLString::transcode(
                                  toolAttrs->getNamedItem(XMLString::transcode("toolID"))->getNodeValue()
                                      )                    );
      // check that the toolID matches the name, but don't make it fatal;
      if (toolID!=m_name) {
         std::cout << "XML toolID does not match this utility" << std::endl;
         std::cout << "toolID= " << toolID << std::endl;
      }
      // Loop over any children of the toolInfo
      if(toolInfo->hasChildNodes()) {
         DOMNodeList* cliArgList = toolInfo->getChildNodes();
         for (unsigned int i=0; i<cliArgList->getLength(); ++i) {
            DOMNode* thisArg = cliArgList->item(i);
            // Only process children that are CliArg's
            if (std::string(XMLString::transcode(thisArg->getNodeName())) == "CliArg") {
               DOMNamedNodeMap* attrs = thisArg->getAttributes();
               // Oy Vay, can we make assigning a string any more complicated
               std::string cla2Name =  std::string(XMLString::transcode(
                                      attrs->getNamedItem(XMLString::transcode("name"))->getNodeValue()
                                              )                    );
               bool exists=false;
               std::vector<CmdLineArgs2*>::const_iterator thisit = this->begin();
               while (thisit != this->end()) {
                  if ((*thisit)->name()==cla2Name) {exists=true; break;}
                  ++thisit;
               }
               if (exists) {
                  std::string option =  std::string(XMLString::transcode(
                                           attrs->getNamedItem(XMLString::transcode("option"))->getNodeValue()
                                                   )                    );
                  args.push_back(option);
                  // Check whether the node has any text associated with it
                  // Treat the first node as that text -- should be true
                  if (thisArg->hasChildNodes()) {
                     DOMNodeList* data = thisArg->getChildNodes(); 
		     if (data->getLength() == 1) { 
			// Old format with single text value for a cliarg 
			std::string optionText =  
			   std::string(XMLString::transcode( 
					  thisArg->getFirstChild()->getNodeValue() 
					  )                    ); 
			//std::cout << "Found text node with value " << optionText << std::endl; 
			//break it apart if it is a multiarg, otherwise take as in quotes text 
			if ( (*thisit)->getArgQual(option).multiarg==false ) { 
			   args.push_back(optionText); 
			} 
			else { 
			   std::vector<std::string> tokens; 
			   Tokenize(optionText,tokens); 
			   for (std::vector<std::string>::iterator tit=tokens.begin(); 
				tit != tokens.end(); 
				++tit) 
			      args.push_back(*tit); 
			}
		     }
                     else {
			// New format with multiple datum elements, each of which is an arg 
			for (unsigned int i=0; i<data->getLength(); ++i) { 
			   if (std::string(XMLString::transcode(data->item(i)->getNodeName())) == "datum") { 
			      std::string argvalue = std::string(XMLString::transcode( 
								    data->item(i)->getFirstChild()->getNodeValue() 
								    )                    ); 
			      args.push_back(argvalue); 
			   } 
			}
		     }
                  }
               }
               else { std::cout << "Class " << cla2Name << " is not used for " 
                                << m_name << std::endl;}
            }
         }
      }
   }
   return args;
}

