#ifndef UTILITIES_COLLECTION_METAINFO
#define UTILITIES_COLLECTION_METAINFO

#include "ArgQual.h"
#include "CmdLineArgs2.h"

/**********************************************************

Metainfo is an concretization of a CmdLineArgs2
   to process metadata argument

Qualifiers: -metadata

**********************************************************/

class MetaInfo : public CmdLineArgs2
{
public:

  /// Constructors
  MetaInfo( );

  virtual ~MetaInfo() {}

  ///
  bool evalArgs(std::vector<std::string>& argv);

  std::pair<std::string,std::string> getMetaEntry(unsigned int);
  std::string           getKeyN(unsigned int entry);
  std::string           getValueN(unsigned int entry);

  size_t                nMeta();

private:
  std::vector< std::string > m_keys;
  std::vector< std::string > m_values;
};


MetaInfo::MetaInfo( ) : CmdLineArgs2("MetaInfo") 
{
   QualList markers;
   markers.insert( std::make_pair("-metadata",       ArgQual(2,true,false)) );
   markers["-metadata"].desc << "<metadata key> <metadata value> " << std::endl
                             << "List of space separated key-value pairs of metadata entries"
                             << "that will be added to the output collection(s)"
                             <<  std::endl;
   setArgQuals(markers);
}


inline bool
MetaInfo::evalArgs(std::vector<std::string>& argv)
{
   if( !CmdLineArgs2::evalArgs(argv) )
      return false;

   if ( this->hasQual("-metadata") ) {
      int ifirst = (*this)["-metadata"].first;
      int ilast  = (*this)["-metadata"].second;
      if( (ilast - ifirst)&1 ) {
         std::cerr << "Metadata key without value" << std::endl;
         return false;
      }
      for( int i = ifirst; i < ilast ; ) {
         m_keys.push_back( argv[i++] );
         m_values.push_back( argv[i++] );
      } 
   } 

   return true;
}


inline std::string
MetaInfo::getKeyN(unsigned int entry)
{
   return m_keys.at(entry);
}

inline std::string
MetaInfo::getValueN(unsigned int entry)
{
   return m_values.at(entry);
}


inline std::pair<std::string,std::string>
MetaInfo::getMetaEntry(unsigned int entry)
{
   return std::pair<std::string,std::string>(m_keys.at(entry),m_values.at(entry));
}


inline size_t  MetaInfo::nMeta()
{
   return m_keys.size();
}

#endif  

