#ifndef UTILITIES_COLLECTION_QUERYINFO
#define UTILITIES_COLLECTION_QUERYINFO

#include "ArgQual.h"
#include "CmdLineArgs2.h"

/**********************************************************

QueryInfo is an concretization of a CmdLineArgs2
   to contain query info.

Qualifiers: -query, -queryopt

**********************************************************/

class QueryInfo : public CmdLineArgs2
{
public:

   /// Constructors
   QueryInfo( );

   virtual ~QueryInfo() {}

   ///
   bool evalArgs(std::vector<std::string>& argv);

   std::string query() {return m_query;}
   std::string queryOptions() {return m_queryOptions;}

private:
   std::string m_query;
   std::string m_queryOptions;

};

QueryInfo::QueryInfo( ) : CmdLineArgs2("QueryInfo") 
{
   QualList markers;
   markers.insert( std::make_pair("-query",     ArgQual()) );
   markers.insert( std::make_pair("-queryopt",  ArgQual()) );
   markers["-query"].desc << "[predicate string (DEFAULT = \"\")]";
   markers["-queryopt"].desc << "<comma separated attribute list> (DEFAULT = * )";
   this->setArgQuals(markers);
}

inline bool
QueryInfo::evalArgs(std::vector<std::string>& argv)
{
   bool retc = true;
   retc = this->CmdLineArgs2::evalArgs(argv);
   if (!retc) return retc;

   if ( this->hasQual("-query") ) {
      int ifirst = (*this)["-query"].first;
      m_query = argv[ifirst];
   }

   if ( this->hasQual("-queryopt") ) {
      int ifirst = (*this)["-queryopt"].first;
      m_queryOptions = argv[ifirst];
   }
        
   return retc;

}


#endif  // UTILITIES_COLLECTION_QUERYINFO

